#include <curl/curl.h>
#include <sys/epoll.h>

#include <stdexcept>
#include <sstream>
#include "shared_ptr.hpp"
#include <errno.h>

#include "curl_fd_leak_test.hh"
extern "C" {
void curl_memdebug(const char *logname);
}
//#include <curlx.h>
//#include <config.h>
//#ifdef CURLDEBUG
//#ifndef CURLTOOLDEBUG
//#define MEMDEBUG_NODEFINES
//#endif
///* This is low-level hard-hacking memory leak tracking and similar. Using
// *    the library level code from this client-side is ugly, but we do this
// *       anyway for convenience. */
//#include "memdebug.h"
//#endif



// finish program after work is done
//#define FINISH 
//#define FIX_FD_LEAK

using namespace std;

static const int NHANDLES = 3;
static const int EPOLL_MAXEVENTS = NHANDLES;
static const int EPOLL_TIMEOUT = 1000; // wait forever
//static const int EPOLL_TIMEOUT = -1; // wait forever


static int epoll_fd=-1; // epoll filedescriptor

ostream& operator<<(ostream& os, const Curl_handles& curl_handles)
{
	os << "running: " << curl_handles.running << endl
		<< "multi: " << hex << (void*) curl_handles.multi << dec << endl;

	for(Curl_handles::handles_t::const_iterator i = curl_handles.handles.begin(); i != curl_handles.handles.end(); ++i ) {
		os << *i->second;
	}
	return os;
}

ostream& operator<<(ostream& os, const Curl_handles::Handle& handle)
{
	os << "easy: " << hex << (void*)handle.easy << dec << endl;
	os << "url: " << handle.doc.url << endl;
	os << "state: ";
	if(handle.state == Curl_handles::UNUSED) 
		os << "UNUSED";
	else if(handle.state == Curl_handles::HEAD)
		os << "HEAD";
	else if(handle.state == Curl_handles::BODY)
		os << "BODY";
	else if(handle.state == Curl_handles::FINISHED)
		os << "FINISHED";
	else 
		os << "other BUG";
	os << endl;	
	return os;		
}

/**
 * Epoll wrappers
 */
int err_sys(std::string s) {
	std::string err;
	err = s + " error: " + strerror(errno) + "\n";
	throw runtime_error(err);
}

int Epoll_ctl(int epfd, int op, int fd, struct epoll_event *event) {
	int n;
	if (  (n=epoll_ctl(epfd,op,fd,event)) < 0 ) {
		if( errno == EEXIST )
			return(n);
		else  {
			std::cout << "errno: " << errno << std::endl;
			err_sys("epoll_ctl");
		}	
	}	
	return n;

}

int Epoll_wait(int epfd, struct epoll_event * events, int maxevents, int timeout) {
	int n;
//again:	
	if (  (n=epoll_wait(epfd,events,maxevents,timeout)) < 0 ) {
		if( errno == EINTR ) {
			std::clog << "epoll_wait interrupted" << std::endl;
			//goto again;
		} else 
			err_sys("epoll_wait");
	}	
	return n;

}

int socket_callback(CURL *easy, curl_socket_t s, int action, void *userp, void *socketp)
{
	struct epoll_event ev;
	bzero(&ev,sizeof(ev));
	DLOG( cout << "socket_callback()" << endl
		<< "easy: " << hex << (void*) easy << dec << endl
		<< "socket: " << (int)s << endl
		<< "action: " << action << endl;)
	switch(action) {
		case CURL_POLL_NONE:
			DLOG(cout << "CURL_NONE: " << s << endl;)	
			break;
		case CURL_POLL_IN:
		case CURL_POLL_OUT:	
		case CURL_POLL_INOUT:
			DLOG(cout << "CURL_POLL_IN/OUT/INOUT: " << s << endl;)	
			ev.events = EPOLLIN|EPOLLET|EPOLLOUT;
			Epoll_ctl(epoll_fd,EPOLL_CTL_ADD,s,&ev);
			break;

		case CURL_POLL_REMOVE:
			DLOG(cout << "CURL_POLL_REMOVE do nothing: " << s << endl;)	
			//ev.events = EPOLLIN|EPOLLET|EPOLLOUT;
			//Epoll_ctl(epoll_fd,EPOLL_CTL_DEL,s,&ev);
			break;

		default:
			throw runtime_error("unknown action");
			break;
	}
	DLOG( cout << endl; )
	return 0;
}

size_t header_write_callback(void *in_ptr, size_t msize, size_t nmemb, Doc* doc) {
	DLOG( cout << "header_write_callback: " << (void*)(doc) << endl << "msize: " << msize << " nmemb: " << nmemb << endl;)
	// curl warrantees one header read at a time
    int result = 0;
    if (doc != NULL) {
		//doc->headers.push_back(string((char*)in_ptr, msize * nmemb));
        result = msize * nmemb;
    }
    return result;
}



size_t write_callback(void *in_ptr, size_t msize, size_t nmemb, Doc* doc) {
	DLOG( cout << "write_callback: " << (void*)(doc) << endl << "msize: " << msize << " nmemb: " << nmemb << endl;)
    int result = 0;
	//DLOG(cout << "write_callback" << (*doc) << endl);
    if (doc != NULL) {
	//	doc->content.append((char*)in_ptr, msize * nmemb);
        result = msize * nmemb;
    }
    return result;
}

Curl_handles::Curl_handles(int num_handles) :
	handles(),
	multi(0),
	running(0)
{
	if(num_handles <= 0)
		throw runtime_error("Curl_handles::Curl_handles Number of handles must be positive integer");

	multi = curl_multi_init();
	for(int i=0; i < num_handles; ++i) {
		boost::shared_ptr<Handle> handle(new Handle());
		handle->easy = curl_easy_init();
		DLOG( cout << "CURLOPT_WRITEFUNCTION: " << (void*)write_callback << " CURLOPT_HEADERFUNCTION: " << (void*)header_write_callback << endl;)
		curl_easy_setopt(handle->easy, CURLOPT_WRITEFUNCTION, write_callback);
		curl_easy_setopt(handle->easy, CURLOPT_HEADERFUNCTION, header_write_callback);
		handles.insert(make_pair(handle->easy,handle));
	}
	//CURLMOPT_PIPELINING
	if( curl_multi_setopt(multi, CURLMOPT_SOCKETFUNCTION, socket_callback) != CURLM_OK)
		throw runtime_error("curl_multi_setopt error");

}

Curl_handles::~Curl_handles()
{
	// remove
	/*
	for(handles_t::iterator i = handles.begin(); i != handles.end(); ++i ) {
		if(curl_multi_remove_handle(multi,i->second->easy) != CURLM_OK)
			cerr << "curl_multi_remove_handle error";
	}	
	*/
	curl_multi_cleanup(multi);	
}




void Curl_handles::process_handles()
{
	DLOG( cout << "process_handles()" << endl;)
	int msgs=0;
	struct CURLMsg *msg=0;
	while((msg = curl_multi_info_read(multi,&msgs)) != NULL) {
		DLOG( cout << "! ! ! !curl msg: " << msg->msg << endl;)
		if(msg->msg == CURLMSG_DONE) {
			// This transfer is finished
			DLOG( cout << (void*) msg->easy_handle << " CURLMSG_DONE" << endl; );
			Curl_handles::handles_t::iterator i;
			// Retrieve Handler Class
			if( (i=handles.find(msg->easy_handle)) != handles.end() ) {
				if(i->first != i->second->easy)
					throw runtime_error("i->first != i->second->easy");
				if(i->second->state == Curl_handles::HEAD) {
					// HEAD request finished
					i->second->state = Curl_handles::BODY;
					DLOG(cout << "HEAD finished... do a GET" << endl << *i->second << endl);

					// Reuse this handle
					/**
					 * curl_multi_remove_handle
					 */
					DLOG(cout << "curl_multi_remove_handle("<< (void*) multi << "," << i->first << ")" << endl;)
					if(curl_multi_remove_handle(multi,i->second->easy) != CURLM_OK)
						throw runtime_error("process_handles curl_multi_add_handle error");

					curl_easy_setopt(i->second->easy, CURLOPT_HTTPGET, 1);

					/**
					 * curl_multi_add_handle
					 */
					DLOG(cout << "curl_multi_add_handle("<< (void*) multi << "," << i->first << ")" << endl;)
					if(curl_multi_add_handle(multi,i->second->easy) != CURLM_OK)
						throw runtime_error("process_handles curl_multi_add_handle error");

				} else if(i->second->state == Curl_handles::BODY) {
					// GET request finished
					i->second->state = Curl_handles::UNUSED; 			
					DLOG(cout << "GET finished" << endl << *i->second);
					DLOG(cout << "curl_multi_remove_handle("<< (void*) multi << "," << i->first << ")" << endl;)
					if(curl_multi_remove_handle(multi,i->second->easy) != CURLM_OK)
						throw runtime_error("process_handles curl_multi_add_handle error");

// Try to avoid FD LEAK
//
#ifdef FIX_FD_LEAK
					DLOG(cout << "curl_easy_cleanup("<< i->second->easy << ")" << endl;)
					boost::shared_ptr<Handle> han = i->second;
					handles.erase(i);
					curl_easy_cleanup(han->easy);
					han->easy = curl_easy_init();
					DLOG(cout << "curl_easy_init("<< han->easy << ")" << endl;)
					handles.insert(make_pair(han->easy,han));

				curl_easy_setopt(handle->easy, CURLOPT_WRITEFUNCTION, write_callback);
				curl_easy_setopt(handle->easy, CURLOPT_HEADERFUNCTION, header_write_callback);

					curl_easy_setopt(han->easy, CURLOPT_URL, han->doc.url.c_str());
					curl_easy_setopt(han->easy, CURLOPT_WRITEDATA, &han->doc); // address of the document
					DLOG( cout << "CURLOPT_WRITEDATA: " << (void*)&han->doc << endl;)
					curl_easy_setopt(han->easy, CURLOPT_HEADERDATA, &han->doc);
					curl_easy_setopt(han->easy, CURLOPT_FILETIME, 1);
					// do a HEAD request first
					//curl_easy_setopt(han->easy, CURLOPT_NOBODY, 1); 
						
					curl_easy_setopt(han->easy, CURLOPT_HTTPGET, 1);

#endif					
				} else
					throw runtime_error("process_handles message for an curl handle with a wrong state");
			} else 
				throw runtime_error("process_handles message for an unregistered curl handle, this is bad");
		}
	}	
	/*
	if( msg == 0 ) {
		DLOG( cout << "no messages" << endl);
	}
	*/

}

void Curl_handles::set_unused_to_work()
{
	for(Curl_handles::handles_t::iterator i = handles.begin(); i != handles.end(); ++i) {
		if( i->second->state == Curl_handles::UNUSED) {
			if( url_queue.empty() ) {
				DLOG( cout << "url_queue empty" << endl;)
				break;
			}
			DLOG( cout << "Curl_handles::UNUSED: " << i->first << endl;)
			//i->second->state = Curl_handles::HEAD;
			i->second->state = Curl_handles::BODY;
			// dequeue next url
			i->second->doc.url = url_queue.front();
			url_queue.pop_front();
			DLOG( cout << "DEQUEUE url: " << i->second->doc.url << " for: " << i->first << endl;)
			// 
			curl_easy_setopt(i->first, CURLOPT_URL, i->second->doc.url.c_str());
			curl_easy_setopt(i->first, CURLOPT_WRITEDATA, &i->second->doc); // address of the document
			DLOG( cout << "CURLOPT_WRITEDATA: " << (void*)&i->second->doc << endl;)
			curl_easy_setopt(i->first, CURLOPT_HEADERDATA, &i->second->doc);
			curl_easy_setopt(i->first, CURLOPT_FILETIME, 1);
			// do a HEAD request first
			//curl_easy_setopt(i->first, CURLOPT_NOBODY, 1); 
				
			curl_easy_setopt(i->first, CURLOPT_HTTPGET, 1);
			DLOG(cout << "curl_multi_add_handle("<< (void*) multi << "," << i->first << ")" << endl;)
			CURLMcode ret = curl_multi_add_handle(multi,i->first);
			if(ret != CURLM_OK) {
				if( ret == CURLM_BAD_EASY_HANDLE )
					cerr << "CURLM_BAD_EASY_HANDLE" << endl;
				cerr << "ret: " << ret << endl;
				throw runtime_error("set_unused_to_work curl_multi_add_handle error");
			}	
		} else {
			DLOG( cout << "Curl_handles::set_unused_to_work busy handler: " << endl << *i->second << endl;)
		}
	}
	 while(CURLM_CALL_MULTI_PERFORM == curl_multi_socket_all(multi, &running)) {
		DLOG( cout << "!!! curl_multi_socket_all running: " << running << endl; )
	}

}

void Curl_handles::wait_events()
{
	int nfds;
	struct epoll_event ep_evs[EPOLL_MAXEVENTS];
	nfds = Epoll_wait(epoll_fd, ep_evs,EPOLL_MAXEVENTS,EPOLL_TIMEOUT);
	for(int i=0; i < nfds; ++i)  {
		//print_events(ep_evs[i].events);
		while(curl_multi_socket(multi, ep_evs[i].data.fd, &running) == CURLM_CALL_MULTI_PERFORM);
	}
}

void Curl_handles::run()
{
	while(true) {

		while(CURLM_CALL_MULTI_PERFORM == curl_multi_socket_all(multi, &running)) {
			DLOG( cout << "curl_multi_socket_all running: " << running << endl; )
		}
		set_unused_to_work();

		long timeout=5;
		if ( curl_multi_timeout(multi,&timeout) == CURLM_OK)
			DLOG( cout << "timeout: " << timeout << endl; )
		else
			cerr << "curl_multi_timeout failure" << endl;

		wait_events();

#ifdef FINISH
		bool busy=false;
		for(Curl_handles::handles_t::iterator i = handles.begin(); i != handles.end(); ++i)
			if( i->second->state != Curl_handles::UNUSED) 
				busy = true;
		if(!busy) {
			DLOG(cout << "We are done!"<<endl;)
			return;
		}	
		busy = false;	
#endif		

		process_handles();

	}
}

int main(int arc, char* argv[]) {
	curl_memdebug("dump");
	if( (epoll_fd = epoll_create(NHANDLES)) < 0 )
		throw runtime_error(strerror(errno));

	Curl_handles curl_handles(NHANDLES); 

	curl_handles.url_queue.push_back("http://www.meneame.net");
	curl_handles.url_queue.push_back("http://curl.haxx.se");
	curl_handles.url_queue.push_back("http://www.linux.org");
	curl_handles.url_queue.push_back("http://www.slashdot.org");
	curl_handles.url_queue.push_back("http://del.icio.us");
	curl_handles.url_queue.push_back("http://www.cisco.com");
	curl_handles.url_queue.push_back("http://www.lwn.net");
	curl_handles.url_queue.push_back("http://www.freshmeat.net");
	curl_handles.url_queue.push_back("http://www.newsforge.com/");
	
//	curl_handles.url_queue.push_back("http://cantresolve/");


	curl_handles.run();

}
